import logging
import os
from typing import Any, List, Tuple, Union

import opendal
from dotenv import load_dotenv
from mcp.server.fastmcp.resources import Resource
from opendal import Entry, Metadata
from opendal.layers import RetryLayer
from pydantic import Field

logger = logging.getLogger("mcp_server_opendal")

# all opendal related environment variables
load_dotenv()
OPENDAL_OPTIONS = {
    k.replace("OPENDAL_", "").lower(): v.lower()
    for k, v in os.environ.items()
    if k.startswith("OPENDAL_")
}


class OpendalResource(Resource):
    """
    OpenDAL Resource provider that handles interactions with different storage services.
    Acts both as a FastMCP Resource and as an interface to OpenDAL operations.

    This resource provider will read the environment variables for the given scheme and use them to configure the opendal operator.

    For example, if the scheme is "mys3", the environment variables should be:

    ```
    OPENDAL_MYS3_TYPE=s3
    OPENDAL_MYS3_BUCKET=mybucket
    OPENDAL_MYS3_REGION=us-east-1
    OPENDAL_MYS3_ENDPOINT=http://localhost:9000
    OPENDAL_MYS3_ACCESS_KEY_ID=myaccesskey
    OPENDAL_MYS3_SECRET_ACCESS_KEY=mysecretkey
    ```
    """

    scheme: str = Field("", description="Storage scheme (e.g., s3, fs)")
    op: Any = None

    def __init__(self, scheme: str):
        scheme = scheme.lower()
        # Configure OpenDAL operator
        opendal_type = OPENDAL_OPTIONS.get(f"{scheme}_type")
        opendal_options = {
            k.replace(f"{scheme}_", ""): v
            for k, v in OPENDAL_OPTIONS.items()
            if k.startswith(f"{scheme}_")
        }
        logger.debug(f"Initializing OpendalResource with options: {opendal_options}")

        # Initialize FastMCP Resource
        super().__init__(
            uri=f"{scheme}://",
            name=f"{scheme} storage",
            description=f"Storage service accessed via OpenDAL {scheme} protocol",
            mime_type="application/vnd.folder",  # for containers/directories
        )

        # Initialize OpenDAL operator
        self.scheme = scheme
        self.op = opendal.AsyncOperator(opendal_type, **opendal_options).layer(
            RetryLayer()
        )
        logger.debug(f"Initialized OpendalResource: {self.op}")

    async def read(self) -> str:
        """
        Read method for scheme-level resources, only returns descriptive information
        Actual file content is obtained through path-specific methods
        """
        info = f"OpenDAL {self.scheme} storage resource.\n\n"
        info += f"To access specific files, use: {self.scheme}://path/to/file\n"
        info += f"To list directory contents, use the 'list' tool with: {self.scheme}://path/to/dir\n"
        return info

    async def list(
        self, prefix: Union[str, os.PathLike], max_keys: int = 1000
    ) -> List[Entry]:
        """List entries with the given prefix"""
        logger.debug(f"Listing entries with prefix: {prefix}")

        if max_keys <= 0:
            return []

        entries = []

        it = await self.op.list(prefix)

        async for entry in it:
            logger.debug(f"Listing entry: {entry}")
            entries.append(entry)
            if len(entries) >= max_keys:
                break

        return entries

    async def read_path(self, path: Union[str, os.PathLike]) -> bytes:
        """Read content from a specific path"""
        logger.debug(f"Reading path: {path}")
        return await self.op.read(path)

    async def stat(self, path: Union[str, os.PathLike]) -> Metadata:
        """Get metadata for a specific path"""
        logger.debug(f"Statting path: {path}")
        return await self.op.stat(path)

    def is_text_file(self, path: Union[str, os.PathLike]) -> bool:
        """Determine if a file is text-based by its extension"""
        text_extensions = {
            ".txt",
            ".log",
            ".json",
            ".xml",
            ".yml",
            ".yaml",
            ".md",
            ".csv",
            ".ini",
            ".conf",
            ".py",
            ".js",
            ".html",
            ".css",
            ".sh",
            ".bash",
            ".cfg",
            ".properties",
        }
        return any(path.lower().endswith(ext) for ext in text_extensions)


def parse_uri(uri: str) -> Tuple[OpendalResource, str]:
    """Parse a URI into a resource and path"""
    from urllib.parse import unquote, urlparse

    logger.debug(f"Parsing URI: {uri}")
    parsed = urlparse(uri)

    scheme = parsed.scheme
    path = parsed.netloc + parsed.path
    path = unquote(path)  # Decode URL-encoded characters
    return (OpendalResource(scheme), path)



import base64
import json
import shutil
import tempfile
from pathlib import Path

import pytest

from mcp_server_opendal.server import (
    OPENDAL_OPTIONS,
    get_info,
    list,
    mcp,
    read,
    register_resources,
)


@pytest.fixture
def test_files():
    """Create a temporary directory with test files"""
    # Create a temporary directory
    temp_dir = tempfile.mkdtemp()

    try:
        # Create some test files and directories
        root_path = Path(temp_dir)

        # Create a text file
        text_file = root_path / "test_text.txt"
        text_file.write_text(
            "This is a test text file\nSecond line content", encoding="utf-8"
        )

        # Create a binary file
        bin_file = root_path / "binary_file.bin"
        bin_file.write_bytes(b"\x00\x01\x02\x03\xff\xfe\xab\xcd")

        # Create a JSON file
        json_file = root_path / "config.json"
        json_data = {"name": "test", "value": 123, "enabled": True}
        json_file.write_text(
            json.dumps(json_data, ensure_ascii=False), encoding="utf-8"
        )

        # Create a subdirectory and a nested file
        subdir = root_path / "subdir"
        subdir.mkdir()
        nested_file = subdir / "nested_file.log"
        nested_file.write_text("Subdirectory file content", encoding="utf-8")

        yield temp_dir
    finally:
        # Clean up the temporary directory
        shutil.rmtree(temp_dir)


@pytest.fixture
def setup_env(test_files, monkeypatch):
    """Set environment variables and OpenDAL configuration"""
    OPENDAL_OPTIONS.clear()
    OPENDAL_OPTIONS.update(
        {
            "fs_type": "fs",
            "fs_root": test_files,
        }
    )

    monkeypatch.setenv("OPENDAL_FS_TYPE", "fs")
    monkeypatch.setenv("OPENDAL_FS_ROOT", test_files)

    # Re-register resources
    register_resources()

    yield


@pytest.mark.asyncio
async def test_list_resources(setup_env):
    """Test listing available resources"""
    resources = await mcp.list_resources()

    assert len(resources) > 0

    resource_schemes = [str(r.uri.scheme) for r in resources if hasattr(r, "uri")]
    assert "fs" in resource_schemes


@pytest.mark.asyncio
async def test_list_directory_contents(setup_env, test_files):
    """Test listing directory contents"""
    result = await list("fs://")

    assert "test_text.txt" in result
    assert "binary_file.bin" in result
    assert "config.json" in result
    assert "subdir" in result

    subdir_result = await list("fs://subdir/")
    assert "nested_file.log" in subdir_result


@pytest.mark.asyncio
async def test_read_text_file(setup_env):
    """Test reading a text file"""
    result = await read("fs://test_text.txt")

    assert "This is a test text file\nSecond line content" in result["content"]
    assert result["mime_type"] in ["text/plain", None]


@pytest.mark.asyncio
async def test_read_binary_file(setup_env):
    """Test reading a binary file"""
    result = await read("fs://binary_file.bin")

    expected_binary = b"\x00\x01\x02\x03\xff\xfe\xab\xcd"
    decoded = base64.b64decode(result["content"])
    assert decoded == expected_binary
    assert result.get("is_binary", False) is True


@pytest.mark.asyncio
async def test_read_json_file(setup_env):
    """Test reading a JSON file"""
    result = await read("fs://config.json")

    assert "test" in result["content"]
    assert result["mime_type"] in ["application/json", "text/plain", None]


@pytest.mark.asyncio
async def test_read_json_file_with_read(setup_env):
    """Test reading a JSON file with read"""
    result = await read("fs://config.json")

    assert "test" in result["content"]
    assert result["mime_type"] in ["application/json", "text/plain", None]


@pytest.mark.asyncio
async def test_get_file_info(setup_env):
    """Test getting file information"""
    result = await get_info("fs://test_text.txt")

    assert "test_text.txt" in result
    content = "This is a test text file\nSecond line content"
    expected_size = len(content.encode("utf-8"))
    assert f"Size: {expected_size} bytes" in result


@pytest.mark.asyncio
async def test_read_nested_file(setup_env):
    """Test reading a nested file"""
    result = await read("fs://subdir/nested_file.log")

    assert "Subdirectory file content" in result["content"]



import logging

from . import server


def main():
    """Main entry point for the package."""
    logging.basicConfig(level=logging.DEBUG)
    server.main()


# Optionally expose other important items at package level
__all__ = ["main", "server"]



import asyncio
import os

from dotenv import load_dotenv
from llama_index.core.agent import ReActAgent, ReActChatFormatter
from llama_index.core.agent.react.prompts import REACT_CHAT_SYSTEM_HEADER
from llama_index.core.llms import ChatMessage
from llama_index.llms.openai_like import OpenAILike
from llama_index.tools.mcp import BasicMCPClient, McpToolSpec

load_dotenv()


# MCP Server Connection Parameters
MCP_HOST = os.getenv("MCP_HOST")
MCP_PORT = os.getenv("MCP_PORT")

SYSTEM_PROMPT = """\
You are an agricultural research assistant. I am researching the impact of climate change on crop yields.
I need you to help me analyze some research data stored in the file system.

Please first list the files in the `fs://data/` directory,
Then you must read the file `README.txt` to understand the data,
then read the necessary CSV files, and summarize the main findings.
"""


async def get_agent(tools: McpToolSpec):
    tools = await tools.to_tool_list_async()
    agent = ReActAgent.from_tools(
        llm=OpenAILike(
            model=os.getenv("OPENAI_MODEL"),
            api_base=os.getenv("OPENAI_ENDPOINT"),
            api_key=os.getenv("OPENAI_API_KEY"),
            is_chat_model=True,
        ),
        tools=list(tools),
        react_chat_formatter=ReActChatFormatter(
            system_header=SYSTEM_PROMPT + "\n" + REACT_CHAT_SYSTEM_HEADER,
        ),
        max_iterations=20,
        verbose=True,
    )
    return agent


async def handle_user_message(message_content: str, agent: ReActAgent):
    user_message = ChatMessage.from_str(role="user", content=message_content)
    response = await agent.achat(message=user_message.content)
    print(response.response)


async def main():
    mcp_tool = McpToolSpec(client=BasicMCPClient(f"http://{MCP_HOST}:{MCP_PORT}/sse"))

    agent = await get_agent(mcp_tool)
    try:
        await handle_user_message("What is the main finding of the data?", agent)
    except Exception as e:
        print(f"Unexpected error: {type(e)}, {e}")


if __name__ == "__main__":
    asyncio.run(main())



import argparse
import base64
import logging
import os
from typing import Any, Dict

from dotenv import load_dotenv
from mcp.server.fastmcp import FastMCP

from mcp_server_opendal.resource import OPENDAL_OPTIONS, OpendalResource, parse_uri

load_dotenv()
default_log_level = os.environ.get("LOG_LEVEL", "INFO").upper()
log_level = getattr(logging, default_log_level, logging.INFO)

# Initialize the FastMCP server - Use the default log level
mcp = FastMCP("opendal_service", log_level=default_log_level)

# Configure logging - still use numeric constants to configure standard logging
logging.basicConfig(level=log_level)
logger = logging.getLogger("mcp_server_opendal")


def register_resources():
    """Register all OpenDAL resources"""
    # Get all available schemes
    schemes = {k.split("_")[0] for k in OPENDAL_OPTIONS.keys()}

    # Clean existing resources and register new resources
    for scheme in schemes:
        try:
            resource = OpendalResource(scheme)
            mcp.add_resource(resource)
            logger.info(f"Registered OpenDAL resource for scheme: {scheme}")
        except Exception as e:
            logger.error(
                f"Failed to register OpenDAL resource for scheme {scheme}: {e}"
            )


# Register resources
register_resources()


# Create a resource template, used to dynamically generate resources
@mcp.resource("{scheme}://{path}")
async def opendal_resource(scheme: str, path: str) -> Dict[str, Any]:
    """
    Access files in OpenDAL service

    Args:
        scheme: storage service scheme
        path: file path

    Returns:
        Dictionary containing file content and metadata
    """
    logger.debug(f"Reading template resource content: {scheme}://{path}")
    try:
        resource = OpendalResource(scheme)
        data = await resource.read_path(path)
        metadata = await resource.stat(path)

        if resource.is_text_file(path):
            return {
                "content": data.decode("utf-8"),
                "mime_type": metadata.content_type or "text/plain",
                "size": metadata.content_length,
                "is_binary": False,
            }
        else:
            return {
                "content": base64.b64encode(data).decode("ascii"),
                "mime_type": metadata.content_type or "application/octet-stream",
                "size": metadata.content_length,
                "is_binary": True,
            }
    except Exception as e:
        logger.error(f"Failed to read resource: {e!s}")
        return {"error": str(e)}


# Modify the list tool to ensure the path ends with a slash
@mcp.tool()
async def list(uri: str) -> str:
    """
    List files in OpenDAL service

    Args:
        uri: resource URI, e.g. mys3://path/to/dir

    Returns:
        String containing directory content
    """
    logger.debug(f"Listing directory content: {uri}")
    try:
        resource, path = parse_uri(uri)

        # Ensure directory path ends with a slash
        if path and not path.endswith("/"):
            path = path + "/"

        entries = await resource.list(path)

        return str(entries)
    except Exception as e:
        logger.error(f"Failed to list directory content: {e!s}")
        return f"Error: {e!s}"


# Read content of file
@mcp.tool()
async def read(uri: str) -> Dict[str, Any]:
    """
    Read file content from OpenDAL service

    Args:
        uri: resource URI, e.g. mys3://path/to/file

    Returns:
        File content or error information
    """
    logger.debug(f"Reading file content: {uri}")
    try:
        resource, path = parse_uri(uri)
        # Directly call the resource function to get content
        return await opendal_resource(resource.scheme, path)
    except Exception as e:
        logger.error(f"Failed to read file content: {e!s}")
        return {"error": str(e)}


# Get file metadata
@mcp.tool()
async def get_info(uri: str) -> str:
    """
    Get metadata of file in OpenDAL service

    Args:
        uri: resource URI, e.g. mys3://path/to/file

    Returns:
        File metadata information
    """
    logger.debug(f"Getting file info: {uri}")
    try:
        resource, path = parse_uri(uri)
        metadata = await resource.stat(path)

        result = f"File: {path}\n"
        result += f"Size: {metadata.content_length} bytes\n"
        result += f"Type: {metadata.content_type}\n"

        return result
    except Exception as e:
        logger.error(f"Failed to get file info: {e!s}")
        return f"Error: {e!s}"


def main():
    parser = argparse.ArgumentParser(description="OpenDAL MCP server")
    parser.add_argument(
        "--transport",
        type=str,
        choices=["stdio", "sse"],
        default="stdio",
        help="Transport method (stdio or sse)",
    )

    args = parser.parse_args()

    if args.transport == "sse":
        mcp.run("sse")
    else:
        mcp.run("stdio")


if __name__ == "__main__":
    main()
